"""Utils for discovering contents, structure, specs of complex objects

>>> x = [
...    (1, 2, 3, 4),
...    {1: range(3), 2: range(2, 40, 2), 5: 6, 7: 8},
...    range(1000),
...    (7, 8, 9, 10),
...    *range(10)
...    ]

>>> head(x)
[(1, 2, 3), {1: (0, 1, 2), 2: (2, 4, 6), 5: 6, 7: 8}, (0, 1, 2)]

>>> head(x, n=2, ellipsis=True)
[(1, 2, ...), {1: (0, 1, ...), 2: (2, 4, ...), ...: ...}, ...]

>>> print(phead(x))
[ (1, 2, 3, ...),
  {1: (0, 1, 2), 2: (2, 4, 6, ...), 5: 6, ...: ...},
  (0, 1, 2, ...),
  ...]

>>> hprint(x, n=5, width=32, indent=4)
[   (1, 2, 3, 4),
    {   1: (0, 1, 2),
        2: (   2,
               4,
               6,
               8,
               10,
               ...),
        5: 6,
        7: 8},
    (0, 1, 2, 3, 4, ...),
    (7, 8, 9, 10),
    0,
    ...]
"""


from pprint import pprint, pformat
from typing import Iterable, Mapping, Sequence, MutableSequence, MutableSet

__all__ = ['head', 'phead', 'hprint']


class Etc:
    """A decorative object to append to truncated iterables.
    """

    def __bool__(self):
        return False

    def __eq__(self, other):
        return other == Ellipsis

    def __repr__(self):
        return '...'

    def __str__(self):
        return '...'

    def __hash__(self):
        return 0

    def __lt__(self, other):
        return False

    def __gt__(self, other):
        return True

    def __cmp__(self, other):
        return 1


def head(x, n=3, ellipsis=False):
    """Recuresivly iterates over a (possibly nested and infinite) iterable, and truncates each
    iterable to a limited length.

    Args:
        x: The input iterable
        n (int, optional): Length of head for each iterable. Defaults to 3.
        ellipsis (bool, optional): Add an ellipsis to the end of each truncated iterator. Defaults to False.

    Returns:
        Iterable: The truncated iterable, with same structure as the input.
    """

    if not isinstance(x, Iterable):
        return x

    if isinstance(x, str):
        return x

    items = []

    if isinstance(x, Mapping):

        for idx, k in enumerate(x):
            if idx == n:
                if ellipsis:
                    items.append((Etc(), Etc()))
                    break
            items.append((k, head(x[k], n, ellipsis)))
        return dict(items)

    else:
        for idx, item in enumerate(x):
            if idx == n:
                if ellipsis:
                    items.append(Etc())
                break
            items.append(head(item, n, ellipsis))
        if isinstance(x, MutableSequence):
            return list(items)
        elif isinstance(x, Sequence):
            return tuple(items)
        elif isinstance(x, MutableSet):
            return set(items)
        else:
            return items


def phead(x, n=3, depth=None, width=80, indent=2, **kwargs):
    """Returns pretty-formatted head of the (possibly nested and infinite) input iterable.
    Args:
        x (int): The input iterable
        n (int, optional): Length of head for each iterable. Defaults to 3.
        depth (int, optional): Max recursion depth. Defaults to None.
        width (int, optional): Max width for each output line. Defaults to 80.
        indent (int, optional): The amount of indentation added for each recursive level. Defaults to 2.

    Returns:
        str: Pretty-formatted head of the input iterable.
    """

    x_head = head(x, n, ellipsis=True)
    return pformat(x_head, depth=depth, width=width, indent=indent, **kwargs)


def hprint(x, n=3, depth=None, width=80, indent=2, **kwargs):
    """Prints prettified head of the (possibly nested and infinite) input iterable.
    Args:
        x (int): The input iterable
        n (int, optional): Length of head for each iterable. Defaults to 3.
        depth (int, optional): Max recursion depth. Defaults to None.
        width (int, optional): Max width for each output line. Defaults to 80.
        indent (int, optional): The amount of indentation added for each recursive level. Defaults to 2.

    Returns:
        Truncated head of the input iterable.
    """

    x_head = head(x, n, ellipsis=True)
    pprint(x_head, depth=depth, width=width, indent=indent, **kwargs)


# def methods(obj, args=None, kwargs=None):
#     ...


# def attributes(obj, types=None):
#     ...


# def parents(obj):
#     ...


# def brief(x):
#     # path, file
#     # function
#     # generator
#     # iterable
#     # custom instance
#     ...


if __name__ == '__main__':

    import doctest
    doctest.testmod()



"""Sorting-related utils.

>>> mylist = [4, 10, -8, 0, 4, 33, 6, 97, 1, 6., 41, -6, 0.0]

>>> topk(mylist, 4)
[97, 41, 33, 10]
>>> topk(mylist, -3)
[-8, -6, 0]

>>> argmax(mylist)
7
>>> argmin(mylist)
2
>>> argsort(mylist)
[2, 11, 3, 12, 8, 0, 4, 6, 9, 1, 5, 10, 7]
>>> argsort(mylist, reverse=True)
[7, 10, 5, 1, 6, 9, 0, 4, 8, 3, 12, 11, 2]
>>> argtopk(mylist, 3)
[7, 10, 5]
>>> argtopk(mylist, 4, reverse=True)
[2, 11, 3, 12]

>>> ascending(mylist)
[-8, -6, 0, 0.0, 1, 4, 4, 6, 6.0, 10, 33, 41, 97]
>>> descending(mylist)
[97, 41, 33, 10, 6, 6.0, 4, 4, 1, 0, 0.0, -6, -8]
>>> reorder(mylist, argsort(mylist))
[-8, -6, 0, 0.0, 1, 4, 4, 6, 6.0, 10, 33, 41, 97]

>>> mydict = {'a': 1, 'b': 4, 'c': -1}

# for dicts, keys are considered as indices
>>> argmin(mydict)
'c'
>>> argmax(mydict)
'b'
>>> argsort(mydict)
['c', 'a', 'b']

>>> reorder(mydict, argsort(mydict))
OrderedDict([('c', -1), ('a', 1), ('b', 4)])
>>> ordered(mydict)
OrderedDict([('c', -1), ('a', 1), ('b', 4)])
>>> ordered(mydict, by_keys=True)
OrderedDict([('a', 1), ('b', 4), ('c', -1)])

"""

from typing import AbstractSet, MutableSequence, MutableSet, AbstractSet, Union, Sequence, Mapping
from collections import OrderedDict
import heapq


__all__ = ['ascending', 'descending', 'topk',
           'argtopk', 'argsort', 'argsorted',
           'argmin', 'argmax', 'reorder', 'ordered']


Sortable = Union[Sequence, Mapping, AbstractSet]


def topk(x: Sortable, k, reverse=False, key=None):
    if reverse or k < 0:
        return heapq.nsmallest(abs(k), x, key=key)
    else:
        return heapq.nlargest(k, x, key=key)


def decide_arg(x: Sortable):
    if isinstance(x, Mapping):
        return x.keys()
    else:
        return range(len(x))


def argtopk(x: Sortable, k, reverse=False):
    return topk(decide_arg(x), k, reverse=reverse, key=x.__getitem__)


def argsort(x: Sortable, reverse=False):
    return sorted(decide_arg(x), key=x.__getitem__, reverse=reverse)


def argsorted(x: Sortable, reverse=False):
    return argsort(x, reverse)


def argmax(x: Sortable):
    return max(decide_arg(x), key=x.__getitem__)


def argmin(x: Sortable):
    return min(decide_arg(x), key=x.__getitem__)


def reorder(x: Sortable, indices):
    if isinstance(x, (Mapping)):
        # dict() is ordered too, by OrderedDict() is more explicit
        return OrderedDict({key: x[key] for key in indices})

    elif isinstance(x, (Sequence)):
        reordered_x = (x[i] for i in indices)
        if isinstance(x, (MutableSequence)):
            return list(reordered_x)
        else:
            return tuple(reordered_x)
    else:
        raise NotImplementedError


def ordered(x: Sortable, by_keys=False, reverse=False):
    if isinstance(x, (Mapping)):
        if by_keys:
            sorted_keys = sorted(x.keys(), reverse=reverse)
        else:
            sorted_keys = argsorted(x, reverse=reverse)
        return reorder(x, sorted_keys)

    elif isinstance(x, (Sequence, AbstractSet)):
        sorted_x = sorted(x, reverse=reverse)
        if isinstance(x, (MutableSequence, MutableSet)):
            return sorted_x
        else:
            return tuple(sorted_x)
    else:
        raise NotImplementedError


def ascending(x, by_keys=False):
    return ordered(x, by_keys=by_keys, reverse=False)


def descending(x, by_keys=False):
    return ordered(x, by_keys=by_keys, reverse=True)


if __name__ == '__main__':

    import doctest
    doctest.testmod()



import csv
from copy import deepcopy
from typing import Any, Dict, List, Union


class Table:
    """Simple pythonic container for tabular data.
    
    # create table
    >>> data = {
    ...    'a': [3, 5, 21638540, 'very-loong-value'],
    ...    'b': [16, 8, 0, 1],
    ...    'c': [0, 5, 6, None],
    ...    }
    >>> table = Table(data)
    >>> print(table.draw())
    a               b    c      
    ============================
    3               16   0      
    5               8    5      
    21638540        0    6      
    very-loong-va.. 1    None
    
    >>> data = [
    ...    {'a': 3, 'b': 16, 'c': 0},
    ...    {'a': 5, 'b': 8, 'c': 5},
    ...    {'a': 21638540, 'b': 0, 'c': 6},
    ...    {'b': 1, 'a': 'very-loong-value'},
    ...    ]
    >>> table = Table(data)
    >>> print(table.draw())
    a               b    c      
    ============================
    3               16   0      
    5               8    5      
    21638540        0    6      
    very-loong-va.. 1    None
    
    # access columns, rows
    >>> table['b']      # access table columns (by str key)
    [16, 8, 0, 1]
    
    >>> table[2]        # access table rows (by int key)
    {'a': 21638540, 'b': 0, 'c': 6}
    
    >>> table[1:3]      # or get a slice of table
    a          b   c   
    ===================
    5          8   5   
    21638540   0   6
    
    # concatenate two tables
    >>> table[:2] + table[2:]
    a               b    c      
    ============================
    3               16   0      
    5               8    5      
    21638540        0    6      
    very-loong-va.. 1    None
    
    # manipulate table columns
    >>> table['d'] = [0] * len(table)
    >>> table
    a               b    c      d   
    ================================
    3               16   0      0   
    5               8    5      0   
    21638540        0    6      0   
    very-loong-va.. 1    None   0
    
    # manipulate table rows
    >>> table[1] = {'a': 5, 'b': 55, 'c': 555}
    >>> table.append({'b': 66, 'c': 666})
    >>> table
    a               b    c      d      
    ===================================
    3               16   0      0      
    5               55   555    None   
    21638540        0    6      0      
    very-loong-va.. 1    None   0      
    None            66   666    None
    
    # csv read/write
    >>> table.to_csv('output.tsv')
    >>> Table.from_csv('output.tsv')
    a               b    c      d      
    ===================================
    3               16   0      0      
    5               55   555    None   
    21638540        0    6      0      
    very-loong-va.. 1    None   0      
    None            66   666    None
    """

    MAX_COL_WIDTH = 16

    def __init__(self, data: Union[List[Dict[str, Any]], Dict[str, List]]) -> None:
        if isinstance(data, dict):
            self.check_cols(data)
            self.cols = data
        elif isinstance(data, list):
            self.cols = {}
            self.cols_from_rows(data)
        else:
            raise ValueError(
                'the `data` must be either a list of dicts (rows) or a dict of lists (cols).')

    @classmethod
    def check_cols(cls, cols):
        lengths = [len(col) for col in cols.values()]
        if len(set(lengths)) != 1:
            raise ValueError(
                f'All table columns must have the same lengths. Column lengths: {lengths}')

    def cols_from_rows(self, rows):
        """Converts list of dicts (rows) to dict of lists (cols), 
        fills missing values with `None`.

        Args:
            rows (List[Dict[str, Any]]): Table rows

        Returns:
            (Dic[str, List]): The dict of lists (cols).
        """
        rows = list(rows)
        col_names = []
        for row in rows:
            col_names.extend(row.keys())

        self.cols = {c: [] for c in dict.fromkeys(col_names)}

        for row in rows:
            self.append(row)

    def row_at(self, index):
        return {name: col[index] for name, col in self.cols.items()}

    def append(self, row):
        for c in self.cols:
            self.cols[c].append(row.get(c))

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.cols[key]
        if isinstance(key, int):
            return self.row_at(key)
        if isinstance(key, slice):
            sliced_data = {c: col[key] for c, col in self.cols.items()}
            return self.__class__(data=sliced_data)
        if isinstance(key, list):
            sliced_data = {c: self.cols[c] for c in key}
            return self.__class__(data=sliced_data)
        raise ValueError(
            'Invalid key. It should be `str` or `List[str]` to access column(s), `int` or `slice` to return row(s).')

    def __setitem__(self, key, value):
        
        if isinstance(key, str):
            if len(value) != len(self):
                raise ValueError(
                    f'New columns values must have the same length with existing cols. {len(value)} != {len(self)}')
            self.cols[key] = value
            
        elif isinstance(key, int):
            if not isinstance(value, dict):
                raise ValueError(f'New row should be a dict with string keys.')
            for c in self.cols:
                self.cols[c][key] = value.get(c)
        else:
            raise ValueError(
                'Invalid key. It should be `str` to set columns or `int` to edit rows.')

    def __len__(self):
        a_col = next(iter(self.cols.values()))
        return len(a_col)

    def __iter__(self):
        for idx in range(len(self)):
            yield self.row_at(idx)

    def draw(self, n_rows=5):

        c2len = {}

        for c, col in self.cols.items():
            lengths = map(len, map(str, col[:n_rows]))
            max_len = max([len(c), *lengths]) + 3
            c2len[c] = min(max_len, self.MAX_COL_WIDTH)

        # header
        result = ''.join([c.ljust(c2len[c]) for c in self.cols]) + '\n'
        result += '=' * sum(c2len.values()) + '\n'

        # rows
        for i, row in enumerate(self):
            if i >= n_rows:
                break
            for c in self.cols:
                str_value = str(row[c])
                if len(str_value) > c2len[c] - 1:
                    str_value = str_value[:c2len[c] - 3] + '..'
                result += str_value.ljust(c2len[c])
            result += '\n'

        return result.strip()

    def __str__(self) -> str:
        return self.draw()

    def __repr__(self) -> str:
        return self.draw()
    
    def copy(self):
        return deepcopy(self)
    
    def __add__(self, other):
        first = self.copy()
        first.extend(other)
        return first
        
    def extend(self, other):
        if type(other) is not type(self):
            try:
                fata = self.__class__(other)
            except:
                raise ValueError('Provide a `Table` or valid data for `Table` constructor.')
        
        if self.cols.keys() != other.cols.keys():
            raise ValueError('The `data` must have the same columns as the source table.')
        
        for c, col in self.cols.items():
            col.extend(other[c])
    
    concat = extend

    def to_csv(self, path, sep=',', header=True, **kwargs):

        with open(path, mode='w', newline='') as csv_file:
            writer = csv.DictWriter(csv_file,
                                    fieldnames=list(self.cols.keys()),
                                    delimiter=sep, **kwargs)
            if header:
                writer.writeheader()
            for row in self:
                writer.writerow(row)
                
    def append_to_csv(self, path, sep=',', **kwargs):

        with open(path, mode='a', newline='') as csv_file:
            writer = csv.DictWriter(csv_file,
                                    fieldnames=list(self.cols.keys()),
                                    delimiter=sep, **kwargs)
            for row in self:
                writer.writerow(row)
                
    @staticmethod
    def from_csv(path, sep=',', header=True,
                 missing_value=None, transform=None, **kwargs):

        cols = {}

        if transform is None:
            def transform(s): return s

        with open(path, mode='r') as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=sep, **kwargs)
            for i, row in enumerate(csv_reader):
                if i == 0:
                    if header:
                        cols.update({c: [] for c in row})
                        continue
                    else:
                        cols.update({i: [] for i in range(len(row))})
                for c, v in zip(cols, row):
                    cols[c].append(transform(v) if v != '' else missing_value)

        return Table(data=cols)


if __name__ == '__main__':
    
    import doctest
    doctest.testmod(verbose=True)




"""Regular expression helper functions.

>>> p = make_regex(
...    following('then '),      # lookbehind
...    grouped(r'[a-z ]+', name='code'),
...    followed_by(r' end'),     # lookahead
...    )
>>> p.pattern
'(?<=then )(?P<code>[a-z ]+)(?= end)'
>>> p.search('if condition then do something end; etc')
<re.Match object; span=(18, 30), match='do something'>

>>> p = make_regex(
...    any_of('ali', 'hossein', 'javad'),
...    SPACES,
...    any_of('goodarzi', 'sharafi', name='family')
...    )
>>> p.pattern
'(?:hossein|javad|ali)\\\\s+(?P<family>goodarzi|sharafi)'
>>> match = p.search('his name is hossein sharafi.')
>>> match
<re.Match object; span=(12, 27), match='hossein sharafi'>
>>> match.group('family')
'sharafi'

>>> p = make_regex(
...    any_of(
...        grouped('<p>', name='p'),
...        grouped('<div>', name='div'),
...        name='tag'
...    ),
...    repeated('.', greedy=False, name='inside'), 
...    if_matched('p', then='</p>', otherwise='</div>')
...    )
>>> p.pattern        # this is really hard to read
'(?P<tag>(?P<div><div>)|(?P<p><p>))(?P<inside>(?:.)+?)(?(p)</p>|</div>)'
>>> text = '<div><p>hello world!</p></div><p>good morning...</p><div>bye!</div>'
>>> for m in p.finditer(text):
...     print(m.group('tag'), m.group('inside'))
<div> <p>hello world!</p>
<p> good morning...
<div> bye!

>>> p = make_regex('[0-9]')
>>> p.sub(repl({'1': 'one', '2': 'two'}), '2 and a half.')
'two and a half.'

>>> p = make_regex('[0-9()\-+/*]{3,}')
>>> p.sub(repl(eval), 'It costs 25*4+10 for you.')
'It costs 110 for you.'
"""

import re
from operator import ior
from typing import Mapping
from functools import reduce


__all__ = ['join_patterns', 'make_regex', 'grouped', 
           'atomic', 'if_matched',
           'followed_by', 'not_followed_by',
           'following', 'not_following', 
           'preceded_by', 'not_preceded_by', 
           'lookahead', 'neg_lookahead',
           'lookbehind', 'neg_lookbehind',
           'repeated', 'possibly',
           'wordbounded', 'any_of', 'charset', 'any_char_but',
           'bag_of_patterns',
           'map2repl', 'func2repl', 'repl', 
           'SPACES', 'DIGITS', 'DOT', 'BOS', 'EOS', 'BOL', 'EOL',
           ]


SPACES = r'\s+'
DIGITS = r'\d+'

DOT = r'\.'
BOS = r'\A'
EOS = r'\Z'
BOL = r'^'
EOL = r'$'


def join_patterns(*patterns, joiner=''):
    if len(patterns) == 1 and not isinstance(patterns[0], str):
        patterns = patterns[0]
    return joiner.join(patterns)


def make_regex(*patterns, ascii=False, debug=False, ignorecase=False,
               locale=False, multiline=False, dotall=False, verbose=False):
    """Compiles the given pattern(s) into a regular expression object, using `re` flags as kwargs.

    Returns:
        re.Pattern: The compiled regex pattern object
    """
    flags = [
        re.ASCII if ascii else 0,
        re.DEBUG if debug else 0,
        re.IGNORECASE if ignorecase else 0,
        re.LOCALE if locale else 0,
        re.MULTILINE if multiline else 0,
        re.DOTALL if dotall else 0,
        re.VERBOSE if verbose else 0]
    return re.compile(join_patterns(patterns), flags=reduce(ior, flags))


def grouped(*pattern, name=None, no_capture=False):
    pattern = join_patterns(pattern)
    if name:
        return fr'(?P<{name}>{pattern})'
    if no_capture:
        return fr'(?:{pattern})'

    return fr'({pattern})'

# special groups

def atomic(*pattern):
    pattern = join_patterns(pattern)
    return fr'(?>{pattern})'


def if_matched(group_name, *, then, otherwise=''):
    return fr'(?({group_name}){then}|{otherwise})'


def followed_by(*pattern, immediately=True):
    pattern = join_patterns(pattern)
    prefix = '' if immediately else '.*'
    return fr'(?={prefix}{pattern})'


def not_followed_by(*pattern, immediately=True):
    pattern = join_patterns(pattern)
    prefix = '' if immediately else '.*'
    return fr'(?!{prefix}{pattern})'


def following(*pattern):
    pattern = join_patterns(pattern)
    return fr'(?<={pattern})'


def not_following(*pattern):
    pattern = join_patterns(pattern)
    return fr'(?<!{pattern})'


preceded_by = following
not_preceded_by = not_following

lookahead = followed_by
neg_lookahead = not_followed_by
lookbehind = preceded_by
neg_lookbehind = not_preceded_by


# extras

def repeated(*pattern, n=None, min=1, max=None, greedy=True, possessive=False, name=None):
    """Repeated pattern

    Args:
        n (int, optional): Exact number of repetitions. If provided, `min`, `max` are ignored. Defaults to None.
        min (int, optional): Minimum number of repetitions. Defaults to 1.
        max (int, optional): Maximum number of repetitions. Defaults to None.
        greedy (bool, optional): Match as much as possible? Defaults to True.
        possessive (bool, optional): Do not backtrack after maching greedily? Defaults to False.
        name (str, optional): The name of pattern group. if None, a non-capturing group is used. Defaults to None.

    """
    if n is not None:
        return grouped(*pattern, name=name, no_capture=True) + fr'{{{n}}}'
    
    min_reps = str(min) if min else ''
    max_reps = str(max) if max else ''
    
    special = ''
    if not greedy:
        special = '?'
    elif possessive:
        special = '+'
    
    repeats = fr'{{{min_reps},{max_reps}}}'
    repeats = repeats.replace('{1,}', '+')
    repeats = repeats.replace('{,}', '*')
    repeats = repeats.replace('{,1}', '?')
    
    pattern = grouped(*pattern, no_capture=True) + fr'{repeats}{special}'
    
    if name:
        return grouped(pattern, name=name)
    return pattern


def possibly(*pattern, name=None):
    return grouped(pattern, name=name, no_capture=True) + '?'


def wordbounded(*pattern, name=None):
    pattern = join_patterns(r'\b', *pattern, r'\b')
    if name:
        return grouped(pattern, name=name)
    return pattern


def any_of(*patterns, sort=True, name=None):
    """Makes the pattern to match any of the input patterns (using `|`). Optionally,
    sorts the patterns starting from the longest to prioritize for the longest. But
    this only works for fixed-length patterns, as the lenght of dynamic patterns 
    cannot be determined.

    Args:
        patterns (Iterable): The input patterns.
        sort (bool, optional): Whether or not to sort the patterns starting from the longest. Defaults to True.
        name (str, optional): The name of pattern group. if None, a non-capturing group is used. Defaults to None.
    """
    if len(patterns) == 1 and not isinstance(patterns[0], str):
        patterns = patterns[0]
    if sort:
        patterns = sorted(patterns, key=len, reverse=True)

    return grouped(r'|'.join(patterns), name=name, no_capture=True)


def charset(characters: str):
    return fr'[{characters}]'


def any_char_but(characters: str):
    return fr'[^{characters}]'


def bag_of_patterns(*patterns):
    """Makes a pattern that matches if all of the input patterns,
    are matched anywhere in the following string.

    Args:
        patterns (Iterable): The input patterns.
    """
    if len(patterns) == 1 and not isinstance(patterns[0], str):
        patterns = patterns[0]
    return ''.join(followed_by(p, immediately=False) for p in patterns)


# repl maker

def func2repl(fn, group=0):
    """Converts a str-to-str function, to a valid `repl` argument for `re.sub()`.

    Args:
        fn (function): A str to str function.
        group (int, optional): The match group to pass to the function. Defaults to 0.
    """
    def repl_func(match):
        return str(fn(match.group(group)))
    return repl_func


def map2repl(mapping: Mapping, group=0, default_str=''):
    """Converts a str-to-str mapping, to a valid `repl` argument for `re.sub()`.

    Args:
        mapping (_type_): _description_
        group (int, optional): The match group to pass to the function. Defaults to 0.
        default_str (str, optional): _description_. Defaults to ''.
    """
    def repl_func(match):
        return str(mapping.get(match.group(group), default_str))
    return repl_func


def repl(rep, group=0):
    if isinstance(rep, Mapping):
        return map2repl(rep, group=group)
    elif callable(rep):
        return func2repl(rep, group=group)
    else:
        raise ValueError(
            'Provide a `Callable` or `Mapping` to make a repl function')


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""Path utils with focus on ease-of-use and method-chaining.
"""

import os
import json
import pickle
import shutil
import pprint
import pathlib
import tempfile
from typing import Any, Dict, List, Union


try:
    from .table import Table
except:
    class Table:
        ...


__all__ = ['Path', 'FilePath', 'TempFilePath']


class Path(str):
    """String-like path with path-related methods, focused on method-chaining.
    Methods return the path-itself when there is no other explicit return value.
    Internally, a `pathlib.Path` is used to invoke most of the operations.

    >>> d = Path('path/to').mkdir()
    >>> p = (d / 'file').with_suffix('.txt').touch()
    >>> p.copy(d / 'copy.txt').name
    'copy.txt'

    >>> len(d.glob('*.txt'))
    2
    >>> d.parent.rmdir(recursive=True).exists()
    False
    """

    def __new__(cls, *args, **kwargs):
        # explicitly only pass value to the str constructor
        path = pathlib.Path(*args)
        return super(Path, cls).__new__(cls, path)

    def __init__(self, *args, **kwargs):
        # ... and don't even call the str initializer
        self._path = pathlib.Path(*args)

    @classmethod
    def cwd(cls):
        """Return a new path pointing to the current working directory
        (as returned by os.getcwd()).
        """
        return cls(pathlib.Path.cwd())

    @classmethod
    def home(cls):
        """Return a new path pointing to the user's home directory (as
        returned by os.path.expanduser('~')).
        """
        return cls(pathlib.Path.home())

    @property
    def name(self):
        """A string representing the final path component, excluding the drive and root, if any"""
        return self.__class__(self._path.name)

    @property
    def suffix(self):
        """The file extension of the final component, if any"""
        return self._path.suffix

    @property
    def stem(self):
        """The final path component, without its suffix"""
        return self._path.stem

    @property
    def parent(self):
        """The logical parent of the path. Note that this is a purely lexical operation"""
        return self.__class__(self._path.parent)

    @property
    def drive(self):
        """A string representing the drive letter or name, if any"""
        return self._path.drive

    @property
    def root(self):
        """A string representing the (local or global) root, if any"""
        return self.__class__(self._path.root)

    @property
    def anchor(self):
        """The concatenation of the drive and root"""
        return self.__class__(self._path.anchor)

    extension = suffix

    # Path manipulation

    def absolute(self):
        """Return an absolute version of this path.  This function works
        even if the path doesn't point to anything.

        No normalization is done, i.e. all '.' and '..' will be kept along.
        Use resolve() to get the canonical path to a file.
        """
        return self.__class__(self._path.absolute())

    def resolve(self, strict=False):
        """
        Make the path absolute, resolving all symlinks on the way and also
        normalizing it (for example turning slashes into backslashes under
        Windows).
        """
        return self.__class__(self._path.resolve(strict))

    def readlink(self):
        """
        Returns the path to which the symbolic link points.
        """
        return self.__class__(self._path.readlink())

    def relative_to(self, other):
        """
        Compute a version of this path relative to the path represented by other.
        """
        return self.__class__(self._path.relative_to(other))

    def join(self, *paths):
        """
        Calling this method is equivalent to combining the path with each of the other arguments in turn.
        """
        return self.__class__(self._path.joinpath(*paths))

    def with_suffix(self, suffix):
        """Returns a new path with the suffix changed. If the original path doesn’t have a suffix,
        the new suffix is appended instead. If the suffix is an empty string, the original suffix is removed"""
        return self.__class__(self._path.with_suffix(suffix))

    def with_name(self, name):
        """Returns a new path with the name changed. If the original path doesn’t have a name, ValueError is raised"""
        return self.__class__(self._path.with_name(name))

    def with_stem(self, name):
        """Returns a new path with the stem changed. If the original path doesn’t have a name, ValueError is raised"""
        return self.__class__(self._path.with_stem(name))

    def with_dir(self, parent_dir):
        """Returns a new path with the parent dir changed."""
        return self.__class__(parent_dir / self.name)

    with_ext = with_suffix

    # Magic methods

    def __truediv__(self, other):
        return self.__class__(self._path / other)

    def __rtruediv__(self, other):
        return self.__class__(other / self._path)

    def __floordiv__(self, other):
        return self.with_dir(other)

    # File/Directory operations

    def touch(self, mode=0o666, exist_ok=True):
        """
        Create this file with the given access mode, if it doesn't exist.
        """
        self._path.touch(mode, exist_ok)
        return self

    def unlink(self, missing_ok=False):     # file or link
        """
        Remove this file or link.
        If the path is a directory, use rmdir() instead.
        """
        self._path.unlink(missing_ok)
        return self

    def rename(self, target, replace=False):
        """
        Rename this path to the target path.

        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.

        Returns the new Path instance pointing to the target path.
        """
        if replace:
            return self.__class__(self._path.replace(target))

        return self.__class__(self._path.rename(target))

    def remove(self, recursive=False, missing_ok=True):
        """Removes both files, directories and even non-empty directory along with its 
        contents (Not default).

        Args:
            recursive (bool, optional): Remove non-empty dir with its contents?. Defaults to False.
            missing_ok (bool, optional): Ignore missing file or dir?. Defaults to True.

        Raises:
            FileNotFoundError: if path does not exist and `missing_ok` is False.

        Returns:
            The path itself.
        """
        if self.is_file():
            return self.unlink()
        elif self.is_dir():
            return self.rmdir(recursive=recursive)
        elif not missing_ok:
            raise FileNotFoundError(
                f'The file or dir `{self}` does not exist!')

    def copy(self, dst, **kwargs):
        """Copy file or directory to the destination.
        If the path is a file `shutil.copy2()` is used, else `shutil.copytree` copies the entire directory.
        """

        if self.is_file():
            return self.__class__(shutil.copy2(self, dst, **kwargs))
        else:
            shutil.copytree(self, dst, **kwargs)

        return self.__class__(dst)

    def rmdir(self, recursive=False):
        """
        Remove this directory.  If the directory is not empty, passing
        `recursive=True` means remove this with its contents.
        """
        if recursive:
            shutil.rmtree(self)
        else:
            self._path.rmdir()
        return self

    def mkdir(self, mode=0o777, parents=True, exist_ok=True):
        """
        Create a new directory at this given path.
        """
        self._path.mkdir(mode, parents, exist_ok)
        return self

    makedirs = mkdir

    # Symbolic link operations

    def symlink_to(self, target, target_is_directory=False):
        """
        Make this path a symlink pointing to the target path.
        Note the order of arguments (link, target) is the reverse of os.symlink.
        """
        self._path.symlink_to(target, target_is_directory)
        return self

    def hardlink_to(self, target):
        """
        Make the target path a hard link pointing to this path.

        Note this function does not make this path a hard link to *target*,
        despite the implication of the function and argument names. The order
        of arguments (target, link) is the reverse of Path.symlink_to, but
        matches that of os.link.
        """
        self._path.link_to(target)
        return self

    symlink = symlink_to
    link = hardlink_to

    # Path queries (neither changes path string nor files, dirs)
    # returns non-Path objects

    def match(self, pattern):
        """Match this path against the provided glob-style pattern"""
        return self._path.match(pattern)

    def same(self, other_path):
        """Return whether other_path is the same or not as this file
        (as returned by os.path.samefile()).
        """
        return self._path.samefile(other_path)

    def chmod(self, mode):
        """
        Change the permissions of the path, like os.chmod().
        """
        self._path.chmod(mode)
        return self

    def lchmod(self, mode):
        """
        Like chmod(), except if the path points to a symlink, the symlink's
        permissions are changed, rather than its target's.
        """
        self._path.lchmod(mode)
        return self

    def exists(self):
        """
        Whether this path exists.
        """
        return self._path.exists()

    def stat(self):
        """
        Return the result of the stat() system call on this path, like
        os.stat() does.
        """
        return self._path.stat()

    def is_dir(self):
        """
        Whether this path is a directory.
        """
        return self._path.is_dir()

    def is_file(self):
        """
        Whether this path is a regular file (also True for symlinks pointing
        to regular files).
        """
        return self._path.is_file()

    def is_mount(self):
        """
        Check if this path is a POSIX mount point
        """
        # Need to exist and be a dir
        return self._path.is_mount()

    def is_symlink(self):
        """
        Whether this path is a symbolic link.
        """
        return self._path.is_symlink()

    def is_block_device(self):
        """
        Whether this path is a block device.
        """
        return self._path.is_block_device()

    def is_char_device(self):
        """
        Whether this path is a character device.
        """
        return self._path.is_char_device()

    def is_fifo(self):
        """
        Whether this path is a FIFO.
        """
        return self._path.is_fifo()

    def is_socket(self):
        """
        Whether this path is a socket.
        """
        return self._path.is_socket()

    def get_size(self, recursive=True):
        """Returns file size in bytes and if the path is a directory,
        (recursively) calculates sum of all file """
        if self.is_file():
            return self.stat().st_size

        file_paths = self.rglob('*') if recursive else self.listdir()

        return sum(f.stat().st_size for f in file_paths if f.is_file())

    matches = match
    isdir = is_dir
    isfile = is_file
    islink = is_symlink
    isfile = is_file
    isfile = is_file

    def owner(self):
        """
        Return the login name of the file owner.
        """
        return self._path.owner()

    def group(self):
        """
        Return the group name of the file gid.
        """
        return self._path.group()

    def open(self, mode='r', buffering=-1, encoding=None,
             errors=None, newline=None):
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return self._path.open(mode, buffering=buffering, encoding=encoding,
                               errors=errors, newline=newline)

    # Directory exploring

    def iterdir(self):
        """Iterates over the child paths in this directory.  Does not 
        yield any result for the special paths '.' and '..'.
        """
        return (self.__class__(p) for p in self._path.iterdir())

    def listdir(self):
        """Lists the child paths in this directory.  Does not yield any
        result for the special paths '.' and '..'.
        """
        return [self.__class__(p) for p in self._path.iterdir()]

    def glob(self, pattern):
        """Iterate over this subtree and yield all existing files (of any
        kind, including directories) matching the given relative pattern.
        """
        return [self.__class__(p) for p in self._path.glob(pattern)]

    def rglob(self, pattern):
        """Recursively yield all existing files (of any kind, including
        directories) matching the given relative pattern, anywhere in
        this subtree.
        """
        return [self.__class__(p) for p in self._path.rglob(pattern)]


class FilePath(Path):
    """Handles file read/write operations which are not relevant to a Path.

    >>> f = FilePath('path/to', 'file.txt')
    >>> f.mkdir().write_text('hello world!').read_text()
    'hello world!'
    >>> f.print({'x': 4, 'y':5}).aprint({'x': 6, 'y':7}).get_size()
    36
    >>> print(f.read_text())
    {'x': 4, 'y': 5}
    {'x': 6, 'y': 7}
    <BLANKLINE>

    >>> f.pickle({'x': 4, 'y': 5}).pickle({'x': 6, 'y': 7}, append=True).unpickle()
    [{'x': 4, 'y': 5}, {'x': 6, 'y': 7}]

    # and so on for `json` and `csv`
    """

    def mkdir(self, mode=0o777, parents=True, exist_ok=True):
        """Make parent directory for the file path.
        """
        Path(self.parent).mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
        return self

    makedirs = mkdir

    def read_bytes(self):
        """
        Open the file in bytes mode, read it, and close the file.
        """
        return self._path.read_bytes()

    def read_text(self, encoding=None, errors=None):
        """
        Open the file in text mode, read it, and close the file.
        """
        return self._path.read_text(encoding, errors)

    def write_bytes(self, data):
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        # type-check for the buffer interface before truncating the file
        self._path.write_bytes(data)
        return self

    def write_text(self, data, encoding=None, errors=None):
        """
        Open the file in text mode, write to it, and close the file.
        """
        self._path.write_text(data, encoding, errors)
        return self

    def readlines(self, n=None) -> List[str]:
        return list(self.ireadlines(n))

    def ireadlines(self, n=None):
        with self.open(mode='r') as f:
            for i, line in enumerate(f):
                if n is not None and i == n:
                    break
                yield line

    def print(self, *objs, pretty=True, **kwargs):
        with open(self, 'w') as f:
            if pretty:
                for obj in objs:
                    pprint.pprint(obj, stream=f, **kwargs)
            else:
                print(*objs, sep='\n', file=f, flush=True)
        return self

    def append_print(self, *objs, pretty=True, **kwargs):
        with open(self, 'a') as f:
            if pretty:
                for obj in objs:
                    pprint.pprint(obj, stream=f, **kwargs)
            else:
                print(*objs, sep='\n', file=f, flush=True)
        return self

    aprint = append_print

    def dump_pickle(self, obj, append=False, **kwargs):
        mode = 'a+b' if append else 'wb'
        with open(self, mode) as f:
            pickle.dump(obj, f, **kwargs)
        return self

    def dump_pickles(self, *objects, append=False, **kwargs):
        mode = 'a+b' if append else 'wb'
        with open(self, mode) as f:
            for obj in objects:
                pickle.dump(obj, f, **kwargs)
        return self

    def iload_pickle(self, **kwargs) -> Any:
        with open(self, 'rb') as f:
            while True:
                try:
                    yield pickle.load(f, **kwargs)
                except (EOFError, IOError):
                    break

    def load_pickle(self, **kwargs) -> Any:
        return list(self.iload_pickle(**kwargs))

    pickle = dump_pickle
    unpickle = load_pickle

    def dump_json(self, obj, **kwargs):
        with open(self, 'w') as f:
            json.dump(obj, f, **kwargs)
        return self

    def load_json(self, **kwargs) -> Any:
        with open(self, 'r') as f:
            return json.load(f, **kwargs)

    def read_csv(self, sep=',', header=True, missing_value=None) -> Table:
        """Read path as csv file and return data as `izy.table.Table`.

        Args:
            sep (str, optional): Separator (delimiter) character. Defaults to ','.
            header (bool, optional): Whether or not to use first row as header. Defaults to True.
            missing_value (Any, optional): The value to replace empty row fileds. Defaults to None.

        Returns:
            izy.table.Table: The csv data
        """
        return Table.from_csv(self, sep=sep, header=header, missing_value=missing_value,)

    def write_csv(self, table: Union[Table, List, Dict], sep=',', header=True):
        """Write tabular data into path as csv file and return the path.

        Args:
            table (Union[Table, List, Dict]): Tabular data. Can be an `izy.table.Table` or 
                any data acceptable for `Table` constructor.
            sep (str, optional): Separator (delimiter) character. Defaults to ','.
            header (bool, optional): Whether or not to use first row as header. Defaults to True.
        """
        if not isinstance(table, Table):
            table = Table(table)
        table.to_csv(self, sep=sep, header=header)
        return self

    def append_csv(self, table, sep=','):
        """Append tabular data into path as csv file and return the path.

        Args:
            table (Union[Table, List, Dict]): Tabular data. Can be an `izy.table.Table` or 
                any data acceptable for `Table` constructor.
            sep (str, optional): Separator (delimiter) character. Defaults to ','.
            header (bool, optional): Whether or not to use first row as header. Defaults to True.
        """
        if not isinstance(table, Table):
            table = Table(table)
        table.append_to_csv(self, sep=sep)
        return self


class TempFilePath(FilePath):
    """Path to a temporary file. If the file is created, it will
    be removed after this path is removed by garbage collector.
    """

    def __new__(cls, dir=None, prefix=None, suffix=None):
        # explicitly only pass value to the str constructor
        fd, path = tempfile.mkstemp(dir=dir, prefix=prefix, suffix=suffix)
        os.close(fd)
        path = pathlib.Path(path)
        return super(Path, cls).__new__(cls, path)

    def __init__(self, *args, **kwargs):
        # ... and don't even call the str initializer
        self._path = pathlib.Path(self)

    def __del__(self):
        if self.exists():
            self.unlink()


if __name__ == '__main__':

    import doctest
    doctest.testmod()



from typing import  Mapping, MutableSequence, Sequence


class Boundict(dict):
    """
    A `Bound` `dict`ionary is bound to a peer which acts as its reverse.
    """
    def __init__(self, peer=None):
        self.peer = peer
        if peer is not None:
            peer.bind(self)
        return super().__init__()

    @staticmethod
    def create_pair():
        a = Boundict()
        b = Boundict(peer=a)
        return a, b

    def bind(self, peer):
        self.peer = peer
        for k, v in self.items():
            super(Boundict, self.peer).__setitem__(v, k)
    
    def __setitem__(self, k, v) -> None:
        if k in self and self[k] != v:
            self.__delitem__(k)
        if v in self.peer and self.peer[v] != k:
            self.peer.__delitem__(v)

        super(Boundict, self).__setitem__(k, v)
        super(Boundict, self.peer).__setitem__(v, k)

    def __delitem__(self, k) -> None:
        super(Boundict, self.peer).__delitem__(self[k])
        super(Boundict, self).__delitem__(k)

    def update(self, other, **kv):        

        new_items = other.items() if isinstance(other, Mapping) else other

        for k, v in new_items:
            self[k] = v
        for k, v in kv.items():
            self[k] = v


class Bidict(Boundict):
    """
    `Bidi`rectional `dict`ionary is an invertible one-to-one mapping (a bijection).
    Obviously both sides should be valid dictionary keys. Uniqueness in either side
    is preserved by removing previous confilctig pairs, without any warnings.

    >>> bd = Bidict()                   # default key names ('direct', 'reverse')
    >>> bd = Bidict('word', 'id')       # custom key names ('word', 'id') 
    >>> bd = Bidict('word-id') 

    >>> bd = Bidict({'salam':0, 'aziam':1, 'khoobi':2})     # default key names
    >>> bd = Bidict(salam=0, aziam=1, khoobi=2)
    >>> bd = Bidict([('salam', 0), ('azizam', 1), ('khoobi', 2)])

    # custom key names
    >>> bd = Bidict(word=['salam', 'azizam', 'khoobi'], id=[0, 1, 2])

    >>> bd['salam'] = 1     # two conflicting pairs ('salam', 0), ('azizam', 1) are removed

    >>> bd.id['salam']      # == bd['salam']
    1
    
    # or more expressive   
    >>> bd.word2id['salam'] 
    1

    >>> bd.word[1]          # == bd.r[1] == bd.id2word[1]
    'salam'
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.r = Boundict(peer=self)

        key_name, value_name = 'reverse', 'direct'

        if len(args) == 1:
            arg = args[0]
            # Bidict('word-id')
            if isinstance(arg, str) and arg.count('-') == 1:
                key_name, _,value_name = arg.partition('-')
            # Bidict({'salam':0, 'aziam':1, 'khoobi':2})
            elif isinstance(arg, Mapping):
                self.update(args[0])
            # Bidict([('salam', 0), ('azizam', 1), ('khoobi', 2)])
            elif isinstance(arg, Sequence):
                self.update(dict(args[0]))
            else:
                raise ValueError('Invalid arguments. Expected one of:'\
                '\nstr containing exactly one "-" e.g. "word-id"'
                '\nMapping'
                '\nSequence'
            )
        # Bidict('word', 'id')
        elif len(args) == 2:
            if all(isinstance(arg, str) for arg in args):
                key_name, value_name = args
            else:
                raise ValueError('Invalid arguments. In case of having two positional arguments, '\
                    'both sould be strings which can be valid variable names e.g. "word", "id".')

        elif len(args) > 2:
            raise ValueError('Invalid arguments. Expected at most two positional arguments.')

        # Bidict(words)
        if len(kwargs) == 2 and all(isinstance(v, MutableSequence) for v in kwargs.values()):
            # MutableSequence is not valid as key/value for bidict
            key_name, value_name = kwargs.keys()
            keys, values = kwargs.values()
            self.update(zip(keys, values))
        else:
            self.update(kwargs)

        setattr(self, 'd', self)
        setattr(self, 'r', self.r)
        setattr(self, value_name, self)         # self.word[*] is supposed to return a word
        setattr(self, key_name, self.r)         # self.id[*] is supposed to return an id

        if key_name != 'direct' and value_name != 'reverse':
            setattr(self, f'{key_name}2{value_name}', self)
            setattr(self, f'{value_name}2{key_name}', self.r)

    def __str__(self) -> str:
        return super().__str__()

    def __repr__(self) -> str:
        dict_repr = super().__repr__()
        return f'{self.__class__.__name__}({dict_repr})'


# class Redict(Bidict):
#     """`Re`versible `dict`ionary is `Bidict` with easier reversed access.
#     Do not use this if direct and reverse values can be equal (e.g. both are numbers). 
#     """
#     def __getitem__(self, k) -> None:
#         if k in self:
#             return super().__getitem__(k)
#         return self.r.__getitem__(k)

#     def __delitem__(self, k) -> None:
#         if k in self:
#             super().__delitem__(k)
#         else:
#             self.r.__delitem__(k)


if __name__ == '__main__':
    
    import doctest
    doctest.testmod(verbose=False)
    


"""Decorator function zoo.
"""

from inspect import isclass
import logging
from time import perf_counter_ns, sleep, perf_counter
import warnings
import functools
from typing import Mapping
from collections import namedtuple
from datetime import datetime
from threading import Thread
from concurrent.futures import ThreadPoolExecutor

from dataclasses import dataclass, field

try:
    from .misc import setup_logger
except:
    def setup_logger(): pass

__all__ = [
    'record_stats', 'register', 'get_function', 'get_class',
    'preprocess', 'postprocess', 'run_before', 'run_after', 
    'retries', 'thread', 'thread_in_pool', 'timeout',
    'returns', 'yields', 'fix_this',
    'ignores', 'fallsback', 'returns_time', 'logs',
]


@dataclass
class UsageStats:
    num_calls: int = 0
    num_errors: int = 0
    total_time: float = 0
    min_time: float = float('inf')
    max_time: float = 0
    error_types: dict = field(default_factory=dict)
    output_types: dict = field(default_factory=dict)
    
    def update_time(self, duration):
        self.total_time += duration
        self.min_time = min(duration, self.min_time)
        self.max_time = max(duration, self.max_time)
        
    def update_errors(self, error):
        self.num_errors += 1
        self.error_types[type(error)] = self.error_types.get(type(error), 0) + 1
        
    def update_outputs(self, output):
        self.output_types[type(output)] = self.output_types.get(type(output), 0) + 1


def record_stats():
    
    stats = UsageStats()
    
    def decorator(fn):
        
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            
            stats.num_calls += 1
            
            try:
                t_before = perf_counter()
                output = fn(*args, **kwargs)
                stats.update_time(perf_counter() - t_before)
                stats.update_outputs(output)
                return output
            except Exception as e:
                stats.update_errors(e)
                raise
        
        wrapper.stats = stats
        return wrapper
    return decorator


# registry --------------------------------------------------------------------

# TODO: make it class-based

function_registry = {}
class_registry = {}


def get_function(name):
    return function_registry[name]

def get_class(name):
    return class_registry[name]


def register(name=None):
    def decorator(func_or_class):
        key = name or func_or_class.__name__
        if isclass(func_or_class):
            class_registry[key] = func_or_class
        else:
            function_registry[key] = func_or_class
        return func_or_class
    return decorator


# chain/pipeline --------------------------------------------------------------------


def chain_functions(*funcs):
    def chained(*args, **kwargs):
        output = funcs[0](*args, **kwargs)
        for func in funcs[1:]:
            output = func(output)
        return output
    return chained

def preprocess(pre_func, *more_funcs):
    def decorator(fn):
        return chain_functions(pre_func, *more_funcs, fn)
    return decorator


def postprocess(post_func, *more_funcs):
    def decorator(fn):
        return chain_functions(fn, post_func, *more_funcs)
    return decorator


# sequence --------------------------------------------------------------------


def run_before(pre_func, *more_funcs):
    def decorator(fn):
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            for func in (pre_func, *more_funcs):
                func(*args, **kwargs)
            return fn(*args, **kwargs)
        return wrapper
    return decorator


def run_after(post_func, *more_funcs):
    def decorator(fn):
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            output = fn(*args, **kwargs)
            for func in (post_func, *more_funcs):
                func(*args, **kwargs)
            return output
        return wrapper
    return decorator


# retry --------------------------------------------------------------------


def retry(n, interval_wait=0, log_level=logging.WARNING):

    def decorator(fn):
        
        ordinal_suffix = {1: 'st', 2: 'nd', 3: 'rd'}
        logger = setup_logger(name=fn.__name__)
        
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            for i in range(1, n+1):
                try:
                    return fn(*args, **kwargs)
                except Exception as e:
                    if i < n:
                        logger.log(
                            log_level,
                            f"{e!r} raised in `{fn.__name__}` in {1}{ordinal_suffix.get(i, 'th')} try out of {n}! Retrying after {interval_wait} seconds..."
                        )
                        sleep(interval_wait)
                    else:
                        raise
        return wrapper
    return decorator


retries = retry


# threading --------------------------------------------------------------------


def thread(deamon=None):
    def decorator(fn):
        
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            thread = Thread(target=fn, name=fn.__name__,
                            args=args, kwargs=kwargs, daemon=deamon)
            thread.start()
            return thread
                
        return wrapper
    
    return decorator

    
def thread_in_pool(max_workers=None):
    
    def decorator(fn):
        
        executor = ThreadPoolExecutor(max_workers=max_workers)
        
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            # returns `Future` object
            return executor.submit(fn, *args, **kwargs)
        
        wrapper.executor = executor
        
        return wrapper
    
    return decorator


def timeout(seconds):
    def decorator(fn):
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            with ThreadPoolExecutor() as executor:
                future = executor.submit(fn, *args, *kwargs)
                return future.result(timeout=seconds)
        return wrapper
    return decorator


# namedtuplify --------------------------------------------------------------------


def returns(*field_names, type_name=None, **name2description):
    def decorator(fn):
        output_type_name = type_name or fn.__name__ + '_output'
        output_field_names = field_names + tuple(name2description.keys())
        FuncOutput = namedtuple(output_type_name, output_field_names)

        doc = fn.__doc__ + '\n' if fn.__doc__ else ''
        doc += 'Returns:\n'
        for name in field_names:
            doc += f'\t{name}\n'
        for name in name2description:
            doc += f'\t{name}: {name2description[name]}\n'
        fn.__doc__ = doc

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            output = fn(*args, **kwargs)
            if isinstance(output, Mapping) and set(output.keys()) == set(field_names):
                return FuncOutput(**output)
            elif isinstance(output, tuple):
                return FuncOutput(*output)
            else:
                return FuncOutput(output)

        return wrapper
    return decorator


def yields(*field_names, type_name=None, **name2description):

    def decorator(fn):
        output_type_name = type_name or fn.__name__ + '_output'
        output_field_names = field_names + tuple(name2description.keys())
        FuncOutput = namedtuple(output_type_name, output_field_names)

        doc = fn.__doc__ + '\n' if fn.__doc__ else ''
        doc += 'Yields:\n'
        for name in field_names:
            doc += f'\t{name}\n'
        for name in name2description:
            doc += f'\t{name}: {name2description[name]}\n'
        fn.__doc__ = doc

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            for item in fn(*args, **kwargs):
                if isinstance(item, Mapping):
                    yield FuncOutput(**item)
                elif isinstance(item, tuple):
                    yield FuncOutput(*item)
                else:
                    yield FuncOutput(item)

        return wrapper
    return decorator


def return_time(milis=False, micros=False, nanos=False):    # defaults to seconds

    def decorator(fn):
        FuncOutput = namedtuple(fn.__name__ + '_output', ('output', 'time'))

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            t_before = perf_counter_ns()
            result = fn(*args, **kwargs)
            duration = perf_counter_ns() - t_before     # nanoseconds
            if milis:
                duration = duration // 1000
            elif micros:
                duration = duration // 1000_000
            elif not nanos: # seconds
                duration = duration / 1000_000_000                

            return FuncOutput(result, duration)

        return wrapper
    return decorator


returns_time = return_time


# logging --------------------------------------------------------------------


def repr_signature(*args, **kwargs):
    args_repr = [repr(a) for a in args]
    kwargs_repr = [f'{k}={v!r}' for k, v in kwargs.items()]
    return ', '.join(args_repr + kwargs_repr)


def repr_call(fn, *args, **kwargs):
    return f'{fn.__name__}({repr_signature(*args, **kwargs)})'


# inspired by https://ankitbko.github.io/blog/2021/04/logging-in-python/
def log(logger=None, to_file=None, file_mode='a',
         before=logging.DEBUG, after=logging.DEBUG, exception=logging.ERROR):
    def decorator(fn):
        logger_name = logger if isinstance(logger, str) else fn.__name__
        mylogger = setup_logger(
            base=logger, name=logger_name, log_file=to_file, mode=file_mode)

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            t_before = datetime.now()
            if before:

                signature = repr_signature(*args, **kwargs)
                mylogger.log(
                    before, f"Function `{fn.__name__}` called with args: ({signature})")
            try:
                result = fn(*args, **kwargs)
                t_after = datetime.now()
                if after:
                    mylogger.log(
                        after, f"Function `{fn.__name__}` returned after {str(t_after-t_before)} with output: {result}")
                return result
            except Exception as e:
                t_exception = datetime.now()
                if exception:
                    mylogger.log(
                        exception, f"{e!r} raised in `{fn.__name__}` after {str(t_exception-t_before)}!\n{str(e)}".strip())
                raise e
        return wrapper
    return decorator


logs = log

# warning --------------------------------------------------------------------


def fix_this(msg=None, warn_each_call=True):
    def decorator(fn):
        message = f'You need to fix `{fn.__name__}`! '
        if msg:
            message += msg
        warnings.warn(message=message)

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            if warn_each_call:
                warnings.warn(message=message)
            return fn(*args, **kwargs)

        return wrapper
    return decorator


# exception --------------------------------------------------------------------


def ignore(*exceptions):
    def decorator(fn):

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except Exception as e:
                for ignored_ex in exceptions:
                    if isinstance(e, ignored_ex):
                        break
                else:
                    raise e

        return wrapper
    return decorator


def fallback(exception, fallback):
    def decorator(fn):

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except exception as e:
                if callable(fallback):
                    return fallback(*args, **kwargs)
                else:
                    return fallback

        return wrapper
    return decorator


ignores = ignore
fallsback = fallback


if __name__ == '__main__':

    @returns('x', 'y')
    def myfunction(x):
        return x, x+1

    @yields('a', 'b')
    def mygenerator(x):
        for i in range(x):
            yield i, i*2

    @fix_this('It always prints ValueError, errors should be raised!')
    def please(x):
        print('ValueError')

    def log_plus(x):
        return x + 1

    def log_minus(s):
        print

    @preprocess(int)
    @postprocess(str)
    def process(x):
        print(type(x))
        return x * 5


    @retry(3)
    def errorenous(x):
        return x / 0
    
    errorenous(10)


from collections.abc import Mapping
import operator as op
import heapq


class Scorer(dict):
    '''Dict subclass for scoring hashable items. Derived from
    builtin python Counter class. Elements are stored as dictionary
    keys and their scores are stored as dictionary values.

    >>> s1 = Scorer({'a': 1, 'b': 2, 'c': 5})
    >>> s1['d'] = 3
    >>> s1
    Scorer({c: 5, d: 3, b: 2, a: 1})

    >>> s2 = Scorer('abc', [-2, 3, 4])
    >>> s2
    Scorer({c: 4, b: 3, a: -2})

    # to support partial add, etc. we carry non-common items from the first scorer unchanged,
    # but those from the second operand are ignored.
    >>> s1 + s2         # binary mathematical operators: +, -, *, /, //, %, **
    Scorer({c: 9, b: 5, d: 3, a: -1})

    >>> s1 / 4          # mathematical operators apply to scalar values
    Scorer({c: 1.25, d: 0.75, b: 0.5, a: 0.25})

    >>> s1 & s2         # ~ min (drops non-common items)
    Scorer({c: 4, b: 2, a: -2})

    >>> s1 & 2          # drops items with score less than 2
    Scorer({c: 5, d: 3, b: 2})

    # ss = (s1 | (s1 & s2)) gives the subset of s1 which also exists in s2
    # you can use this to force math operations to return common items only
    >>> s1 | s2         # ~ max (also aplicable to scalars)
    Scorer({c: 5, b: 3, d: 3, a: 1})

    >>> abs(s2)         # overriden functions: abs(), round()
    Scorer({c: 4, b: 3, a: 2})

    >>> +s2             # special unary oprator (drops items with scores <= 0)
    Scorer({c: 4, b: 3})

    >>> s1 + (+s2)      # not equivalent to s1 + s2
    Scorer({c: 9, b: 5, d: 3, a: 1})
    '''

    def __init__(self, *args, missing_value=0, **kwds):
        '''Initialize a Scorer object with optionally given values.
        '''
        self.missing_value = missing_value
        if len(args) > 2:
            raise TypeError('expected at most 1 arguments, got %d' % len(args))
        super(Scorer, self).__init__()
        self.update(*args, **kwds)

    def __missing__(self, key):
        return self.missing_value

    def bottomk(self, k):
        if k in [None, 0, float('inf')]:
            return self.topk(float('-inf'))
        return self.topk(-k)

    def middlek(self, k):
        raise NotImplementedError

    def best(self):
        return max(self.items(), key=op.itemgetter(1))
    
    def worst(self):
        return min(self.items(), key=op.itemgetter(1))

    def ascending(self):
        return sorted(self.items(), key=op.itemgetter(1), reverse=False)

    def descending(self):
        return sorted(self.items(), key=op.itemgetter(1), reverse=True)

    def median(self):
        if len(self) == 0:
            return None
        med_idx = (len(self)+1) // 2
        return self.topk(med_idx + 1)[med_idx]

    def topk(self, k=1):
        '''List the top k items and their scores.
        If k is negative, then list the bottom k worst items ~ `bottomk(abs(k))`.
        elif k is None, zero or inf, then list all items starting from the top ~ `descending()`,
        elif k is -inf, then list all items starting from the bottom ~ `ascending()`.
        '''
        if k in [None, 0, float('inf')]:
            return self.descending()
        if k == float('-inf'):
            return self.ascending()
        if k < 0:
            return heapq.nsmallest(abs(k), self.items(), key=op.itemgetter(1))
        else:
            return heapq.nlargest(k, self.items(), key=op.itemgetter(1))

    # Override dict methods where necessary

    def update(self, *args, **kwds):
        '''Like dict.update() but add scores instead of replacing them.
        Source can be an iterable, a dictionary, or another Scorer instance.
        '''

        if len(args) > 2:
            raise TypeError('expected at most 2 arguments, got %d' % len(args))
        # print(args)
        if len(args) == 2:
            new_scores = dict(zip(args[0], args[1]))
        elif len(args) == 1:
            new_scores = dict(args[0])  # expected list of tuples
        else:
            new_scores = None
        
        if new_scores is not None:
            if self:
                for elem, score in new_scores.iteritems():
                    self[elem] = self[elem] + score
            else:
                super(Scorer, self).update(new_scores)      # fast path when scorer is empty
        if kwds:
            self.update(kwds)

    def subtract(self, other):
        return self - other

    def copy(self):
        'Return a shallow copy.'
        return self.__class__(self)

    def __reduce__(self):
        return self.__class__, (dict(self),)

    def __delitem__(self, elem):
        'Like dict.__delitem__() but does not raise KeyError for missing values.'
        if elem in self:
            super(Scorer, self).__delitem__(elem)

    def __repr__(self):
        if not self:
            return f'{self.__class__.__name__}()'
        max_num = 9
        space = ' '
        if len(self) <= max_num:
            items = f',{space}'.join(map(lambda x: f'{x[0]}: {x[1]}', self.topk(k=max_num)))
        if len(self) > max_num:
            top = f',{space}'.join(map(lambda x: f'{x[0]}: {x[1]}', self.topk(k=max_num//2)))
            median =  '{}: {}'.format(*self.median())
            bottom = f',{space}'.join(map(lambda x: f'{x[0]}: {x[1]}', self.topk(k=-(max_num//2))))
            items = f'{top},{space}...,{space}{median},{space}..,{space}{bottom}'
        return f'{self.__class__.__name__}({{{items}}})'

    def _binary_operator(self, other, operator):
        result = self.copy()
        if isinstance(other, Mapping):
            for elem, score in self.items():  # not efficient!
                if elem in other:
                    result[elem] = operator(score, other[elem])
        elif isinstance(other, (int, float)):
            for elem, score in self.items():
                result[elem] = operator(score, other)
        else:
            raise ValueError
        return result

    def _unary_operator(self, operator):
        result = Scorer()
        for elem, score in self.items():
            result[elem] = operator(score)
        return result

    def __pos__(self):
        result = Scorer()
        for elem, score in self.items():
            if score > 0:
                result[elem] = score
        return result

    def __neg__(self):
        return self._unary_operator(op.__neg__)
    def __abs__(self):
        return self._unary_operator(op.__abs__)

    def __add__(self, other):
        return self._binary_operator(other, op.__add__)
    def __sub__(self, other):
        return self._binary_operator(other, op.__sub__)
    def __round__(self, other):
        return self._binary_operator(other, round)

    def __mul__(self, other):
        return self._binary_operator(other, op.__mul__)
    def __truediv__(self, other):
        return self._binary_operator(other, op.__truediv__)
    def __floordiv__(self, other):
        return self._binary_operator(other, op.__floordiv__)
    def __pow__(self, other):
        return self._binary_operator(other, op.__pow__)
    def __mod__(self, other):
        return self._binary_operator(other, op.__mod__)

    def __or__(self, other):
        '''Maximum of value in either of the Scorers.
        The other value can be a scalar with the same behavior.
        '''
        result = self.copy()
        if isinstance(other, Scorer):
            for elem, score in other.items():
                if elem in self:
                    result[elem] = max(self[elem], score)
                else:
                    result[elem] = score
        elif isinstance(other, (int, float)):
            for elem, score in self.items():
                result[elem] = max(score, other)
        else:
            raise NotImplementedError

        return result

    def __and__(self, other):
        '''Minimum of corresponding scores among two Scorers.
        If other is a scalar, drops items with score < other.
        '''
        result = Scorer()
        if isinstance(other, Scorer):
            for elem, score in self.items():
                if elem in other:
                    result[elem] = min(score, other[elem])
        elif isinstance(other, (int, float)):
            for elem, score in self.items():
                if score >= other:
                    result[elem] = score
        else:
            raise NotImplementedError
        return result


if __name__ == '__main__':
    
    import doctest
    doctest.testmod()


"""Miscellaneous utility functions and classes
"""

import sys
import logging
import operator
import functools

__all__ = ['Singleton', 
           'flatten', 'without_dups', 'unique', 'first',
           'setup_logger']


class Singleton(type):
    """Singleton metaclass.

    copy-pasted from https://stackoverflow.com/q/6760685/7798781

    >>> class MyClass(metaclass=Singleton):
    ...    pass
    
    >>> x = MyClass()
    >>> x is MyClass()
    True
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(
                Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def flatten(iterable2d):
    """Convert a 2D nested iterable into a flat `list`.

    The approach is claimed to be one of the most efficient options. (https://stackoverflow.com/a/45323085)

    Args:
        iterable2d (iterable): The 2D iterable

    Returns:
        list: The flat list
    """
    return functools.reduce(operator.iconcat, iterable2d, [])


def without_dups(sequence):
    """Returns a copy of the given sequence without duplicate values while preserving order.
    """
    sequence.__class__(dict.fromkeys(sequence))


unique = without_dups


def first(iterable, default=None, condition=None):
    """
    Returns the first item in the `iterable` that satisfies the `condition`.
    If the condition is not given, returns the first item of the iterable.

    If the iterable is empty or no item matches the `condition`,
    the `default` argument is returned.

    simplified version of https://stackoverflow.com/a/35513376/7798781

    Args:
        iterable (iterable): The input iterable :)
        default (_type_, optional): If the iterable is empty or no item matches
            the `condition`, this value is returned. Defaults to None.
        condition (callable, optional): The condition function. Defaults to None.

    Raises:
        ValueError: If the `default` value does not satidfy the provided `condition`

    Returns:
        The first item in the `iterable` that satisfies the `condition` or the
        `default` value.
    """
    try:
        return next(x for x in iterable if condition is None or condition(x))
    except StopIteration:
        if condition is None or condition(default):
            return default
    raise ValueError('The `default` value does not satidfy the provided `condition`')


def setup_logger(base=None, name=None, log_file=None, mode='w'):
    """Initiates a simple logger with a general log format and an optional file handler.

    Args:
        base (logging.Logger, optional): Base logger. Defaults to None.
        name (str, optional): Name of the new logger, 
            if base logger is not provided. Defaults to None.
        log_file (str, optional): Path to the log file. Defaults to None.
        mode (str, optional): Log file mode. Defaults to 'w'.

    Returns:
        logging.Logger: The logger object
    """
    formatter = logging.Formatter(fmt='[{asctime}][{name}][{levelname}] - {message}',
                                  datefmt='%Y-%m-%d %H:%M:%S', style='{')
    logger = base or logging.getLogger(name)
    if log_file:
        file_handler = logging.FileHandler(log_file, mode=mode)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    if base is None:
        screen_handler = logging.StreamHandler(stream=sys.stdout)
        screen_handler.setFormatter(formatter)
        logger.addHandler(screen_handler)
    return logger


if __name__ == '__main__':

    import doctest
    doctest.testmod()
